#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <pthread.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <errno.h>

#include "fb_serv.h"

#define DEBUG 0

pthread_mutex_t mutex;
pthread_t thread_players[MAX_PLAYERS];
int cur_thread_players_nb;

int server_state = SERVER_STATE_READY;
state_player_t tab_state_players[MAX_PLAYERS];

void init_tab_state_players(void)
{
        int i, j, k;

        for (i = 0; i < MAX_PLAYERS; i++) {
                for(j = 0; j < BOARD_WIDTH; j++) {
                        for(k = 0; k < BOARD_HEIGHT; k++) {
                                tab_state_players[i].tab_balls[j][k]=0;
                        }
                }
                tab_state_players[i].state_game = 0;
                tab_state_players[i].player_id  = NOPLAYER;
                tab_state_players[i].ready      = 0;
                strcpy(tab_state_players[i].nickname, "No Player");
        }
}


void write_to_socket(int sock_fd, char * message)
{
        size_t count;
        ssize_t len_write;

#if DEBUG
        printf("write_to_socket %d\n\"%s", sock_fd, message);
#endif
        count = strlen ( message );
        len_write = write(sock_fd, message, count + 1);
        if (len_write != count + 1) {
                if (len_write < 1) {
                        perror("write_to_socket");
                        exit(EXIT_FAILURE);
                }                
        }
}


void write_to_all_socket_except_one(int sock_fd, char * message)
{
        size_t count;
        ssize_t len_write;
        int player_sock_fd;
        int i;

#if DEBUG
        printf("write_to_all_socket_except_one %d \n\"%s", sock_fd, message);
#endif
        count = strlen ( message );
        for (i = 0; i < MAX_PLAYERS; i++) {
                if (tab_state_players[i].player_id != NOPLAYER) {
                        player_sock_fd = tab_state_players[i].sock_fd;
                        if (player_sock_fd != sock_fd) {
                                len_write = write(player_sock_fd, message, count + 1);
                                if (len_write != count + 1) {
                                        if (len_write < 1) {
                                                perror("write_to_all_socket_except_one");
                                                exit(EXIT_FAILURE);
                                        }
                                }
                        }
                }
        }
}


void write_to_all_socket(char * message)
{
        size_t count;
        ssize_t len_write;
        int player_sock_fd;
        int i;

#if DEBUG
        printf("write_to_all_socket\n\"%s", message);
#endif
        count = strlen ( message );
        for (i = 0; i < MAX_PLAYERS; i++) {
                if (tab_state_players[i].player_id != NOPLAYER) {
                        player_sock_fd = tab_state_players[i].sock_fd;
                        len_write = write(player_sock_fd, message, count+1);
                        if (len_write != count + 1) {
                                if (len_write < 1) {
                                        perror("write_to_all_socket");
                                        exit(EXIT_FAILURE);
                                }
                        }
                }
        }
}


void list_players(int sock_fd)
{
        char buffer_out[BUFFER_SIZE];
        int i;
        
        for (i = 0; i < MAX_PLAYERS; i++) {
                if (tab_state_players[i].player_id != NOPLAYER) {
                        strcpy(buffer_out, "");
                        sprintf(buffer_out,
                                "player %d : %s %s\n",
                                tab_state_players[i].player_id,
                                tab_state_players[i].nickname,
                                tab_state_players[i].ready ? "Ready" : "Not ready"
                                );
                        write_to_socket(sock_fd, buffer_out);
                }
        }
}

int get_free_id (void)
{
        int i;
        for (i = 0; i < MAX_PLAYERS; i++) {
                if (tab_state_players[i].player_id == NOPLAYER)
                        return i;
        }
        fprintf(stderr, "No place for new id\n");
        exit(EXIT_FAILURE);
}

void *thread_player_manager(void *args)
{
        int i;
        char buffer_in[BUFFER_SIZE];
        char buffer_out[BUFFER_SIZE];
        int len_read;
        int sock_fd;
        int player_id = -1;
        int is_everybody_ready;
        int nickname_already_in_use;

        sock_fd= *(int *)(args);

#if DEBUG
        printf("cur_thread_players_nb %d --> ", cur_thread_players_nb);
#endif
        pthread_mutex_lock(&mutex);
        cur_thread_players_nb++;
        pthread_mutex_unlock(&mutex);
#if DEBUG
        printf("%d\n", cur_thread_players_nb);
#endif

        if (server_state != SERVER_STATE_ACCEPTING_PLAYERS) {
                write_to_socket(sock_fd, MSG_SERVER_FULL);
                goto end_thread;
        }

        player_id = get_free_id();
        tab_state_players[player_id].player_id = player_id;
        tab_state_players[player_id].sock_fd = sock_fd;
        sprintf(tab_state_players[player_id].nickname, "Player%d", player_id);
        strcpy(buffer_out, COPYRIGHT HOMEPAGE MSG_WELCOME);
        strcat(buffer_out, tab_state_players[player_id].nickname);
        strcat(buffer_out, " !\n");
        strcat(buffer_out, MSG_HELP);
        if (player_id == 0) {
                strcat(buffer_out, MSG_WELCOME_MASTER);
        }
        write_to_socket(sock_fd, buffer_out);

        strcpy(buffer_out, "");
        sprintf(buffer_out, "--> %s has joined the game\n", tab_state_players[player_id].nickname);
        write_to_all_socket(buffer_out);

        do {
                memset(buffer_in, '\0', BUFFER_SIZE);
                len_read = read(sock_fd,buffer_in,BUFFER_SIZE);
                if (len_read > 0) {
                        /* if this is only a cariage return or smthing else, ignore it */
                        if (strlen(buffer_in) < 3) {
                                write_to_socket(sock_fd, "Received a CR, ignoring it...\n");
/*                                 continue; */
                        }
/*                         buffer_in[strlen(buffer_in)-2] = '\0'; */
                        if (strncmp(buffer_in, "", 3) == 0) {
/*                                 printf("Received a Ctrl-C, abording ...\n"); */
                                write_to_socket(sock_fd, "Received a Ctrl-C, abording ...\n");
/*                                 continue; */
                                break;
                        }

#if DEBUG
                        printf("Parsing %d bytes : \"%s\"\n", strlen(buffer_in), buffer_in);
#endif
                        if (strstr(buffer_in, LIST_CMD) == buffer_in) {
                                if (player_id == 0) {
#if DEBUG
                                        printf("pthread_mutex_lock\n");
#endif
                                        pthread_mutex_lock(&mutex);
                                        list_players(sock_fd);
#if DEBUG
                                        printf("pthread_mutex_unlock\n");
#endif
                                        pthread_mutex_unlock(&mutex);
                                } else {
                                        write_to_socket(sock_fd, MSG_NOT_MASTER);
                                }
                        } else 
                        if (strstr(buffer_in, QUIT_CMD) == buffer_in) {
                                break;
                        } else
                        if (buffer_in[0] != '/') {
                                strcpy(buffer_out,"");
                                sprintf(buffer_out, "<%s> %s\n", tab_state_players[player_id].nickname, buffer_in );
                                write_to_all_socket(buffer_out);
                        } else
                        switch (server_state) {

                        case SERVER_STATE_ACCEPTING_PLAYERS: {
                                if (strstr(buffer_in, NICK_CMD) == buffer_in) {
                                        /* Check if the nick doesn't exist */
                                        nickname_already_in_use = 0;
                                        for (i=0; i<MAX_PLAYERS; i++) {
                                                if (i != player_id &&
                                                   strcmp(tab_state_players[i].nickname, buffer_in+strlen(NICK_CMD)) == 0
                                                   ) {
                                                        nickname_already_in_use = 1;
                                                }
                                        }
                                        if (!nickname_already_in_use) {
                                                strcpy(buffer_out, "");
                                                sprintf(buffer_out, 
                                                        "--- %s is known as %s ---\n",
                                                        tab_state_players[player_id].nickname,
                                                        buffer_in+strlen(NICK_CMD)
                                                        );
                                                write_to_all_socket(buffer_out);
#if DEBUG
                                                printf("pthread_mutex_lock\n");
#endif
                                                pthread_mutex_lock(&mutex);
                                                strcpy(tab_state_players[player_id].nickname, buffer_in+strlen(NICK_CMD));
#if DEBUG
                                                printf("pthread_mutex_unlock\n");
#endif
                                                pthread_mutex_unlock(&mutex);
                                        } else {
                                                write_to_socket(sock_fd, "Nickname already in use, please choose another one\n");
                                        }
                                } else {
                                        if (strstr(buffer_in, NEW_CMD) == buffer_in) {
                                                if (player_id == 0) {
                                                        if (cur_thread_players_nb >= 2) {
#if DEBUG
                                                                printf("pthread_mutex_lock\n");
#endif
                                                                pthread_mutex_lock(&mutex);
                                                                server_state = SERVER_STATE_START_GAME;
                                                                /* Don't mark as ready the master */
/*                                                                 tab_state_players[player_id].ready = 1; */
#if DEBUG
                                                                printf("pthread_mutex_unlock\n");
#endif
                                                                pthread_mutex_unlock(&mutex);
                                                                write_to_all_socket(MSG_TYPE_START);
                                                        } else {
                                                                write_to_socket(sock_fd, "Really want to play alone ??\n");
                                                        }
                                                } else {
                                                        write_to_socket(sock_fd, MSG_NOT_MASTER);
                                                }
                                        } else {
                                                write_to_socket(sock_fd, MSG_INVALID_CMD);
                                        }
                                } 
                                break;
                        }       
                        case SERVER_STATE_START_GAME: {
                                if (strstr(buffer_in, START_CMD) == buffer_in) {
#if DEBUG
                                        printf("pthread_mutex_lock\n");
#endif
                                        pthread_mutex_lock(&mutex);
                                        tab_state_players[player_id].ready = 1;
#if DEBUG
                                        printf("pthread_mutex_unlock\n");
#endif
                                        pthread_mutex_unlock(&mutex);

                                        is_everybody_ready = 1;
                                        for (i=0; i < MAX_PLAYERS; i++) {
                                                if (tab_state_players[i].player_id != NOPLAYER) {
                                                        if (tab_state_players[i].ready != 1) {
                                                                write_to_socket(sock_fd, MSG_WAITING_EVERYBODY);
                                                                is_everybody_ready = 0;
                                                                break;
                                                        }
                                                }
                                        }
                                        
                                        if (is_everybody_ready) {
#if DEBUG
                                                printf("pthread_mutex_lock\n");
#endif
                                                pthread_mutex_lock(&mutex);
                                                server_state = SERVER_STATE_PLAY;
                                                write_to_all_socket(MSG_GAME_START);
#if DEBUG
                                                printf("pthread_mutex_unlock\n");
#endif
                                                pthread_mutex_unlock(&mutex);
                                        }
                                }
                                else {
                                        write_to_socket(sock_fd, MSG_TYPE_START);
                                }
                                break;
                        }

                        case SERVER_STATE_PLAY: {
                                if (strstr(buffer_in, PLAYER_STATUS_CMD) == buffer_in
                                    || strstr(buffer_in, PLAYER_MALUS_CMD) == buffer_in
                                    || strstr(buffer_in, PLAYER_SYNCH_CMD) == buffer_in
                                    || strstr(buffer_in, PLAYER_BOARD_CMD) == buffer_in) {
                                        write_to_all_socket_except_one(sock_fd, buffer_in);
                                } else {
                                        write_to_socket(sock_fd, MSG_INVALID_CMD);
                                }
                                break;
                        }

                        default:
                                write_to_socket(sock_fd, MSG_INVALID_CMD);
                        }
                }
        }
        while (len_read > 1);

        strcpy(buffer_out, "");
        sprintf(buffer_out, "<-- %s has left the game\n", tab_state_players[player_id].nickname);
        write_to_all_socket(buffer_out);
        // don't update player_id if this is the master, to prevent reconnexion of a new master.
        tab_state_players[player_id].player_id = NOPLAYER;
        if (player_id == 0) {
                write_to_all_socket_except_one(sock_fd, "Master has quit, game canceled\n");
                pthread_mutex_lock(&mutex);
                server_state = SERVER_STATE_ACCEPTING_PLAYERS;
                pthread_mutex_unlock(&mutex);
        }
        
 end_thread:
        printf("Closing socket\n");
        write_to_socket(sock_fd, "Server is closing socket, bye !\n");
        close(sock_fd);

#if DEBUG
        printf("cur_thread_players_nb %d --> ", cur_thread_players_nb);
#endif
        pthread_mutex_lock(&mutex);
        cur_thread_players_nb--;
        pthread_mutex_unlock(&mutex);
#if DEBUG
        printf("%d\n", cur_thread_players_nb);
#endif

        if (pthread_cancel(pthread_self()) != 0) {
                fprintf(stderr,"cancel error : %d\n", (long int) pthread_self());
        }

        pthread_exit(NULL);
        if (pthread_detach(pthread_self()) != 0) {
                fprintf(stderr,"detach error : %d\n", (long int) pthread_self());
        }        
}



void * thread_serv(void * args)
{
        int sock_serv;
        int sock_client;
        pthread_attr_t thread_attr;
        int optval;
        struct sockaddr_in nom;

	printf(COPYRIGHT HOMEPAGE);
        init_tab_state_players();
        server_state = SERVER_STATE_ACCEPTING_PLAYERS;

        if ((sock_serv = socket(AF_INET, SOCK_STREAM, 0)) <= 0) {
                perror("socket");
                exit(EXIT_FAILURE);
        }
        
        if (setsockopt(sock_serv, SOL_SOCKET, SO_REUSEADDR, &optval, sizeof(int)) != 0) {
                perror("setsockopt");
                exit(EXIT_FAILURE);
        }
        
        bzero((char *)&nom, sizeof(nom));
        
        //init for bind call
        nom.sin_family = AF_INET;
        nom.sin_port=htons(SERVER_PORT);
        nom.sin_addr.s_addr = INADDR_ANY;
        
        if (bind(sock_serv,(struct sockaddr *)&nom, sizeof(nom)) == -1) {
                perror("bind");
                exit(EXIT_FAILURE);
        }
        
        if (listen(sock_serv, MAX_PLAYERS) < 0) {
                perror("listen");
                exit(EXIT_FAILURE);
        }
        
        //if all is ok, launch server thread status.
        //pthread_create(&pth_status,NULL,&service_status,(void *)NULL);
	printf("Server listening\n");

        pthread_mutex_init(&mutex,NULL); 

        while (1) {
                if ((sock_client = accept(sock_serv,NULL,0)) < 0) {
                        perror("accept"); 
                        exit(EXIT_FAILURE);
                } 
                                
                if (pthread_attr_init (&thread_attr) != 0) {
                        perror("pthread_attr_init");
                        exit (EXIT_FAILURE); 
                } 
                
                if (pthread_attr_setdetachstate (&thread_attr, PTHREAD_CREATE_DETACHED) != 0) {
                        perror("pthread_attr_setdetachstate error");
                        exit (EXIT_FAILURE);
                }
                                
                if (pthread_create(&thread_players[get_free_id()],
                                   &thread_attr,
                                    thread_player_manager,
                          (void *) &sock_client) != 0) {
                        perror("pthread_create");
			exit(EXIT_FAILURE);
                }
        }
}

  
int main()
{
        thread_serv(NULL);
        exit(EXIT_SUCCESS);
}
